# This file is part of the Tomographer project, which is distributed under the
# terms of the MIT license.
#
# The MIT License (MIT)
#
# Copyright (c) 2016 ETH Zurich, Institute for Theoretical Physics, Philippe Faist
# Copyright (c) 2017 Caltech, Institute for Quantum Information and Matter, Philippe Faist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function

import codecs
import os
import os.path
import sys
import subprocess
import re
import glob
import shutil



# include path is set in setup.py -- this is the tomographer.include sub-package ...
from include import ensure_str, BOOST_DEPS_COMPONENTS   # beurk!!


def get_version_info(version):

    # major/minor sections of version
    m = re.match(r'^v(?P<major>\d+)\.(?P<minor>\d+)', version)
    version_maj = int(m.group('major'))
    version_min = int(m.group('minor'))

    # Normalize version string for PIP/setuptools
    version_for_pip = version
    # remove initial 'v' in 'v3.1'
    if version_for_pip[0] == 'v':
        version_for_pip = version_for_pip[1:]
    # make PEP-440 compatible if it is a specific git-describe commit number
    m = re.match(r'^(?P<vtag>.*)-(?P<ncommits>\d+)-(?P<githash>g[a-fA-F0-9]+)$', version_for_pip)
    if m:
        version_for_pip = "{vtag}+git{ncommits}.{githash}".format(**m.groupdict())

    return (version_maj, version_min, version_for_pip)



def setup_sources(thisdir, vv):

    if not os.path.exists(os.path.join(thisdir, '..', 'tomographer', 'tomographer_version.h.in')):
        # we use ../tomographer/tomographer_version.h.in to see if we are being run in an
        # original Tomographer source checkout (not necessarily git, but full project
        # checked out as opposed to a python sdist pacakge)
        print("Running from Python packaged source")
        return False

    # We are being run from checked out Tomographer project sources.  Prepare the python
    # sources correctly.

    #
    # setup the VERSION file
    #
    version = None
    try:
        version = ensure_str(subprocess.check_output([vv.get('GIT'), 'describe', '--tags', 'HEAD'],
                                                     cwd=thisdir, stderr=subprocess.STDOUT)).strip()
    except Exception:
        pass
    if not version:
        if os.path.exists(os.path.join(thisdir, '..', 'VERSION')):
            with open(os.path.join(thisdir, '..', 'VERSION')) as f:
                version = ensure_str(f.read()).strip()
    if not version:
        raise RuntimeError("Cannot determine Tomographer version (no git info and no VERSION file)")

    # create VERSION file here -- for sdist
    with open(os.path.join(thisdir, 'VERSION'), 'w') as f:
        f.write(version + "\n")

    (version_maj, version_min, version_for_pip) = get_version_info(version)


    #
    # Include our source headers in package data, along with dependencies
    #


    class ignore_not_header(object):
        def __init__(self, thisdir, but_keep=[]):
            self.thisdir = thisdir
            self.but_keep = but_keep

        def __call__(self, d, files):
            return [ f
                     for f in files
                     if (not os.path.isdir(os.path.join(self.thisdir,d,f)) and not f.endswith('.h') and not f.endswith('.hpp')
                         and not f in self.but_keep) ]

    #
    # Copy all tomographer headers inside tomographer.include as package data
    #
    src_tomographer_include = os.path.join(thisdir, '..', 'tomographer')
    target_tomographer_include = os.path.join(thisdir, 'tomographer', 'include')
    if os.path.exists(os.path.join(target_tomographer_include, 'tomographer')):
        shutil.rmtree(os.path.join(target_tomographer_include, 'tomographer'))
    shutil.copytree(src_tomographer_include,
                    os.path.join(target_tomographer_include, 'tomographer'),
                    ignore=ignore_not_header(thisdir))

    #
    # Copy all tomographerpy headers inside tomographer.include as package data
    #
    src_tomographerpy_include = os.path.join(thisdir, 'cxx', 'tomographerpy')
    target_tomographer_include_tomographerpy = os.path.join(thisdir, 'tomographer', 'include', 'tomographerpy')
    if os.path.exists(os.path.join(target_tomographer_include_tomographerpy)):
        shutil.rmtree(os.path.join(target_tomographer_include_tomographerpy))
    shutil.copytree(src_tomographerpy_include,
                    os.path.join(target_tomographer_include_tomographerpy),
                    ignore=ignore_not_header(thisdir))

    #
    # Copy all dependency headers (boost, eigen)
    #
    target_tomographer_include_deps = os.path.join(target_tomographer_include, 'deps')
    if not os.path.exists(target_tomographer_include_deps):
        os.mkdir(target_tomographer_include_deps)

    if (not vv.get('Boost_INCLUDE_DIR') or
        os.path.realpath(vv.get('Boost_INCLUDE_DIR')).startswith(os.path.realpath(target_tomographer_include_deps))):
        raise ValueError("Invalid Boost_INCLUDE_DIR: "+repr(vv.get('Boost_INCLUDE_DIR')))

    target_tomographer_include_deps_boost = os.path.join(target_tomographer_include_deps, 'boost')
    if os.path.exists(target_tomographer_include_deps_boost):
        shutil.rmtree(target_tomographer_include_deps_boost)
    os.mkdir(target_tomographer_include_deps_boost)
    # Arrgh! "/usr/bin/bcp --boost=/usr/include ..." will copy all system files as well!!
    # solution: create a symlink to '/usr/include/boost' in a sandboxed directory...
    if not os.path.isdir(os.path.join(thisdir,'tmp')):
        os.mkdir(os.path.join(thisdir,'tmp'))
    if not os.path.isdir(os.path.join(thisdir,'tmp','boost_dir')):
        os.mkdir(os.path.join(thisdir,'tmp','boost_dir'))
    if not os.path.exists(os.path.join(thisdir,'tmp','boost_dir','boost')):
        os.symlink(os.path.join(vv.get('Boost_INCLUDE_DIR'), 'boost'), # source
                   os.path.join(thisdir,'tmp','boost_dir','boost')) # link_name

    subprocess.check_output([vv.get('BCP'), '--boost='+os.path.join(thisdir,'tmp','boost_dir')] +
                             BOOST_DEPS_COMPONENTS +
                             [target_tomographer_include_deps_boost ])

    if (not vv.get('EIGEN3_INCLUDE_DIR') or
        os.path.realpath(vv.get('EIGEN3_INCLUDE_DIR')).startswith(os.path.realpath(target_tomographer_include_deps))):
        raise ValueError("Invalid EIGEN3_INCLUDE_DIR: "+repr(vv.get('EIGEN3_INCLUDE_DIR')))

    target_tomographer_include_deps_eigen3 = os.path.join(target_tomographer_include_deps, 'eigen3')
    if os.path.exists(target_tomographer_include_deps_eigen3):
        shutil.rmtree(target_tomographer_include_deps_eigen3)
    os.mkdir(target_tomographer_include_deps_eigen3)
    shutil.copytree(os.path.join(vv.get('EIGEN3_INCLUDE_DIR'), 'Eigen'),
                    os.path.join(target_tomographer_include_deps_eigen3, 'Eigen'),
                    ) # no ignore! Public headers don't have an extension, e.g. "#include <Eigen/Core>"
    shutil.copytree(os.path.join(vv.get('EIGEN3_INCLUDE_DIR'), 'unsupported'),
                    os.path.join(target_tomographer_include_deps_eigen3, 'unsupported'),
                    ) # no ignore! Public headers don't have an extension, e.g. "#include <Eigen/Core>"
    shutil.copy2(os.path.join(vv.get('EIGEN3_INCLUDE_DIR'), 'signature_of_eigen3_matrix_library'),
                 target_tomographer_include_deps_eigen3)
    
    #
    # create tomographer_version.h
    #
    tomographer_version_h_content = """\
/* This file is part of the Tomographer project, which is distributed under the
 * terms of the MIT license.
 *
 * The MIT License (MIT)
 *
 * Copyright (c) 2016 ETH Zurich, Institute for Theoretical Physics, Philippe Faist
 * Copyright (c) 2017 Caltech, Institute for Quantum Information and Matter, Philippe Faist
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef zzzTOMOGRAPHER_VERSION_H_
#define zzzTOMOGRAPHER_VERSION_H_

#define TOMOGRAPHER_VERSION "{version}"
#define TOMOGRAPHER_VERSION_MAJ {version_maj}
#define TOMOGRAPHER_VERSION_MIN {version_min}

#endif
""".format(version=version,version_maj=version_maj,version_min=version_min)
    with open(os.path.join(thisdir, 'tomographer', 'include', 'tomographer', 'tomographer_version.h'), 'w') as f:
        f.write(tomographer_version_h_content)

    #
    # include LICENSE file in this directory
    #
    shutil.copy2(os.path.join(thisdir, '..', 'LICENSE.txt'), thisdir)

    #
    # create the README file
    #
    NOTE_PKG_DEPS = """
Note: This package contains a subset of the
`Boost library <https://boost.org/>`_
(distributed under the `Boost software license <http://www.boost.org/users/license.html>`_)
and the `Eigen3 library <https://eigen.tuxfamily.org/>`_
(distributed under the `MPL license 2.0 <https://www.mozilla.org/en-US/MPL/2.0/>`_).
They are located in the source package directory ``tomographer/include/deps/``.
"""

    readme_content = ''
    with open(os.path.join(thisdir, 'README_.rst')) as f:
        readme_content = f.read()
    with open(os.path.join(thisdir, 'README.rst'), 'w') as fw:
        fw.write(readme_content)
        fw.write(NOTE_PKG_DEPS)

    
    #
    # All set.
    #
    return True




"""
Utilities for fitting the histogram data and computing the *quantum error bars*.
"""

from __future__ import print_function

import collections # namedtuple
import inspect
import numpy as np
import scipy
import scipy.optimize as sciopt
import matplotlib.pyplot as matplt

import logging
logger=logging.getLogger(__name__)


class _Ns:
    pass




def fit_fn_default(x, a2, a1, m, c):
    """
    The default fit model for the logarithm of the histogram data.  The `x`-values are not
    directly the figure of merit, but may be possibly transformed via a `f`-to-`x`
    transformation (see :py:class:`HistogramAnalysis`).

    Returns the value :math:`-a_2\\,x^2 - a_1\\,x + m\\log(x) + c`, which is the default
    fit model.  If a `NumPy` vector is specified as `x`, the function is evaluated for all
    the given values and the values are returned as a vector.
    """
    return -a2*np.square(x) - a1*x + m*np.log(x) + c

QuantumErrorBars = collections.namedtuple('QuantumErrorBars', ('f0', 'Delta', 'gamma') )


def fit_histogram(normalized_histogram, fit_fn, ftox, **kwopts):
    """
    Fit the histogram data to a model function.

    :param normalized_histogram: the final histogram data, provided as a
        :py:class:`tomographer.HistogramWithErrorBars` class (or a subclass
        thereof), and normalized to a proper probability density (see
        :py:meth:`tomographer.HistogramWithErrorBars.normalized()`).

    :param ftox: a function which transforms the figure of merit values (`f`-values) into
        the natural parameter for the fit function (`x`-values).  For example, for the
        fidelity, one should use :math:`x = 1 - f`.

    :param threshold_fraction: a fraction (i.e. value between 0 and 1) of the maximum peak
        value over which the data points are kept; values below the threshold are ignored
        for the fit. (By default, `threshold_fraction=0` and all points are considered;
        since the error bars are taken into account this should be fine in most cases.)

    :param kwopts: additional keyword options are passed on to
        `scipy.optimize.curve_fit(...)`.
    """

    f = normalized_histogram.values_center
    x = ftox(f)
    p = normalized_histogram.bins
    errp = normalized_histogram.delta

    #print(f)
    #print(p)
    #print(errp)

    # Optionally: filter and keep only the points which are above a threshold_fraction of
    # the max of the prob curve.  Indeed, low points are usually not very useful.
    #
    # Points are not filted out by default (threshold_fraction=0)
    thresfrac = kwopts.pop('threshold_fraction', 0)
    thres = thresfrac*np.max(p)
    #display(thres)

    # get the indexes in the histogram where the points are selected (above threshold)
    idxok = np.where(p>thres)

    # form data series of those points we keep
    fok = f[idxok]
    xok = x[idxok]
    logpok = np.log(p[idxok])
    errlogpok = np.divide(errp[idxok], p[idxok]);

    # and perform the fit
    popt, pcov = sciopt.curve_fit(fit_fn, xok, logpok, sigma=errlogpok, absolute_sigma=True, **kwopts)
    #display(popt)
    #display(pcov)
    
    # return value: an object with useful values stored as properties
    d = _Ns()
    d.f = f
    d.x = x
    d.p = p
    d.errp = errp
    d.normalized_histogram = normalized_histogram
    d.idxok = idxok
    d.fok = fok
    d.xok = xok
    d.logpok = logpok
    d.errlogpok = errlogpok
    # fit data
    d.popt = popt
    d.pcov = pcov

    return d



def deskew_logmu_curve(a2, a1, m, c):
    """
    De-skew the fit model to obtain a second order approximation at the peak maximum (see
    details of how to calculate the quantum error bars in our paper).

    Usage::

        (a, x0, y0) = deskew_logmu_curve(a2, a1, m, c)
    """
    
    x0 = (np.sqrt(np.square(a1) + 8*a2*m) - a1) / (4*a2)
    y0 = -a2*np.square(x0) - a1*x0 + m*np.log(x0) + c
    a = a2 + m / (2 * np.square(x0))

    return (a, x0, y0)


def reskew_logmu_curve(m, a, x0, y0):
    a2 = a - m / (2 * np.square(x0))
    a1 = (a2*m - 2*np.square(a2)*np.square(x0)) / (a2*x0)
    c = y0 + a2*np.square(x0) + a1*x0 - m*np.log(x0)
    return (a2, a1, m, c)


def qu_error_bars_from_deskewed(xtof, m, a, x0, y0):
    f0 = xtof(x0)
    Delta = 1/np.sqrt(a)
    gamma = m / (6 * np.square(a) * np.power(x0,3))
    return (f0, Delta, gamma)


def qu_error_bars_to_deskewed_c(ftox, f0, Delta, gamma, y0=1):
    a = 1.0 / np.square(Delta)
    x0 = ftox(f0)
    m = gamma * 6 * np.square(a) * np.power(x0,3)
    return (m, a, x0, y0)


def calc_redchi2(x, y, dy, theo_fn, num_fit_params):

    redchi2 = np.sum( [
        np.square( ( y[k] - theo_fn(x[k]) ) / dy[k] )
        for k in range(len(x))
    ])
    redchi2 /= (len(x) - num_fit_params)

    return redchi2


def guess_fitparams_from_data(histogram, ftox_fn):
    # read off graphically x0, y0, Delta, and gamma from the data -- just a rough guess
    # with absolutely no garantees
    
    idxmax = np.argmax(histogram.bins)

    fvals = histogram.values_center
    xvals = np.array([ ftox_fn(f) for f in fvals ])
    pvals = histogram.bins

    x0 = xvals[idxmax]
    f0 = fvals[idxmax]
    expy0 = histogram.bins[idxmax]
    y0 = np.log(expy0)

    #logger.debug("x0={},f0={},expy0={},y0={}".format(x0,f0,expy0,y0))

    # find values above 1/e relative height of y0
    thresval = expy0*np.exp(-1.0)

    # linearly interpolate left intersect from bordering data points
    firstidxabove = np.argmax(histogram.bins > thresval)
    if firstidxabove == 0: firstidxabove = 1
    x0m = ( xvals[firstidxabove-1] +
            (thresval-histogram.bins[firstidxabove-1])/(histogram.bins[firstidxabove]-histogram.bins[firstidxabove-1])
            *(xvals[firstidxabove]-xvals[firstidxabove-1]) )

    lastidxabove = histogram.numBins() - np.argmax(histogram.bins[::-1] > thresval) - 1
    if lastidxabove+1 >= histogram.numBins(): lastidxabove -= 1
    x0p = ( xvals[lastidxabove+1] +
            (thresval-histogram.bins[lastidxabove+1])/(histogram.bins[lastidxabove]-histogram.bins[lastidxabove+1])
            *(xvals[lastidxabove]-xvals[lastidxabove+1]) )
    
    Delta = np.absolute((x0p - x0m) / 2.0)
    gamma = 2*np.absolute((x0p + x0m) / 2.0 - x0)

    #logger.debug("xvals={}".format(xvals))
    #logger.debug("firstidxabove={},lastidxabove={},x0m={},x0p={},f0={},Delta={},gamma={}".format(firstidxabove,lastidxabove,x0m,x0p,f0,Delta,gamma))

    (a2,a1,m,c) = reskew_logmu_curve(*qu_error_bars_to_deskewed_c(ftox_fn, f0, Delta, gamma, y0=y0))

    if a2 < 0 or m < 0:
        logger.info("Having trouble guessing fit parameters graphically, resorting to generic guess")
        return (500.0, 100.0, 20., 0)

    return (a2, a1, m, c)


class HistogramAnalysis(object):
    """
    Take care of analyzing histogram data obtained from performing a random walk over
    state space according to the distribution :math:`\\mu_{B^n}(\\cdot)` defined by the
    tomography data, while collecting a histogram of a figure of merit.

    Arguments to the constructor:
      - `final_histogram`: the final histogram returned by the random walks
        procedure.  It is expected to be a
        `tomographer.HistogramWithErrorBars`.
      - `ftox`: Specify how to transform the figure of merit value `f` into
        the `x` coordinate for fit. Specify the transformation as a pair of
        values `(h, s)` in the relation `x = s*(f-h)` or `f=s*x+h`, where h
        can be any constant, and where `s` must be plus or minus one. By
        default there is no transformation (`x=f`, corresponding to
        `ftox=(0,1)`.  For the fidelity, you should use `x = 1-f`
        (`ftox=(1,-1)`). For an entanglement witness, you might use `x = 2-f`
        (`ftox=(2,-1)`) for example.
      - `fit_fn`: a function which serves as fit model (for the log of the data)
      - additional named arguments in `kwopts` are passed on to :py:func:`fit_histogram`.
    """
    def __init__(self, final_histogram, ftox=(0,1), fit_fn=None, **kwopts):
        self.final_histogram = final_histogram
        self.normalized_histogram = final_histogram.normalized()
        if ftox[1] not in [1, -1]:
            raise ValueError("Invalid value of `s` in `ftox=(h,s)`: s=%r"%(ftox[1]))
        self.ftox_hs = ftox

        if fit_fn:
            # User specified custom fit function
            self.custom_fit_fn = True
            self.fit_fn = fit_fn
        else:
            self.custom_fit_fn = False
            self.fit_fn = fit_fn_default
            if 'bounds' not in kwopts:
                # a2, a1, m, c
                kwopts['bounds'] = ( (0, -np.inf, 0, -np.inf), np.inf, )
            if 'p0' not in kwopts:
                kwopts['p0'] = guess_fitparams_from_data(self.normalized_histogram, self.ftox)
                logger.debug("Guessing initial fit params = {!r}".format(kwopts['p0']))

        self.FitParamsType = collections.namedtuple('FitParamsType', inspect.getargspec(self.fit_fn).args[1:])

        self.fit_histogram_result = fit_histogram(self.normalized_histogram, fit_fn=self.fit_fn,
                                                  ftox=self.ftox, **kwopts)
        self.fit_params = self.FitParamsType(*self.fit_histogram_result.popt)
        self.fit_params_cov = self.FitParamsType(*np.diagonal(self.fit_histogram_result.pcov))

        self.printFitParameters(print_func=logger.debug,show_cov=True)

        # calculate reduced chi2 statistic
        xok = self.fit_histogram_result.xok
        logpok = self.fit_histogram_result.logpok
        errlogpok = self.fit_histogram_result.errlogpok

        self.redchi2 = calc_redchi2(xok, logpok, errlogpok,
                                    lambda x: self.fit_fn(x, *self.fit_params), num_fit_params=4)
        
        redchi2_warn_threshold = kwopts.get('redchi2_warn_threshold', 1.4)

        if self.redchi2 > redchi2_warn_threshold :
            logger.warning(("Reduced chi-squared statistic = {:.4g}. It could be that the fit "
                            "model isn't good or that the fit optimization failed.").format(self.redchi2))

    
    def xtof(self, x):
        """
        Convert an `x`-value (natural variable of the fit model) to the value of the figure of
        merit (`f`).
        """
        return self.ftox_hs[1]*x + self.ftox_hs[0]

    def ftox(self, f):
        """
        Convert an `f`-value (value of the figure of merit) to an `x`-value (natural variable
        of the fit model).
        """
        return self.ftox_hs[1]*(f - self.ftox_hs[0])

    def fitParameters(self):
        """
        Return the parameters of the fit.  The return value type is a named tuple whose fields
        are the argument names of the fit model `fit_fn` specified to the constructor.
        For example, for the default fit model :py:func:`fit_fn_default`, the fields are
        `(a2, a1, m, c)`.
        """
        return self.fit_params

    def fitRedChi2(self):
        """
        Return the `Reduced chi-squraed statistic
        <https://en.wikipedia.org/wiki/Reduced_chi-squared_statistic>`_ for the fit.
        Should be close to `1` for a good fit.
        """
        return self.redchi2

    def printFitParameters(self, print_func=print, show_cov=False):
        """
        Display the fit parameters.

        By default the values are displayed using `print(...)`. You may specify a custom
        print function if you want them displayed somewhere else.

        The fit parameters are returned in the same format as :py:meth:`fitParameters()`.
        """
        def fmt_param_val(x, cov):
            if show_cov:
                return '{:>8g}  +-  {:.3g}'.format(x, np.sqrt(cov))
            else:
                return '{:g}'.format(x)
        print_func("Fit parameters:\n" + "\n".join([
            '{:>12s} = {}'.format(k, fmt_param_val(getattr(self.fit_params, k), getattr(self.fit_params_cov, k)))
            for k in self.fit_params._fields
        ]))
        return self.fit_params

    def quantumErrorBars(self):
        """
        Calculate the quantum error bars.

        This function is only available if you use the default fit model (i.e., if you
        didn't specify the `fit_fn` argument to the constructor).

        Returns a :py:class:`QuantumErrorBars` named tuple.
        """
        
        if self.custom_fit_fn:
            raise RuntimeError("You cannot call quantumErrorBars() when you specify a custom fit model.")

        (f0, Delta, gamma) = qu_error_bars_from_deskewed(
            self.xtof,
            self.fit_params.m,
            *deskew_logmu_curve(*self.fit_params)
        )
        return QuantumErrorBars(f0, Delta, gamma)

    def printQuantumErrorBars(self, print_func=print):
        """
        Calculates and displays the values of the quantum error bars.

        The quantum error bars are returned, provided in the same format as
        :py:meth:`quantumErrorBars()`.
        """
        q = self.quantumErrorBars()
        print_func(("Quantum Error Bars:\n"+
                    "          f0 = {f0:.4g}\n"+
                    "       Delta = {Delta:.4g}\n"+
                    "       gamma = {gamma:.4g}\n").format(
            f0=q.f0,
            Delta=q.Delta,
            gamma=q.gamma))
        return q


    def plot(self,
             log_scale=False,
             xlabel='Distribution of values',
             plot_deskewed_gaussian=True,
             show_plot=True,
             curve_npts=200,
             **kwopts):
        """
        Plot the histogram data using `matplotlib`.

        :param log_scale: if `True`, then use a logarithmic scale for the y-axis.

        :param xlabel: the label for the x axis (i.e., the figure of merit). Note this
            is the `f`-value, not the `x`-value.

        :param show_plot: if `True` (the default), then the plot is displayed immediately.
            If `False`, then the returned object has an additional method `show()` which can
            be used to display the plot.
        
        :param curve_npts: Number of points for the smooth fit curve plot.

        The return value is an object with the attributes `fig` (the `matplotlib` figure
        object) and `ax` (the `matplotlib` axes object) which you can use to set specific
        custom properties.  If you specified `show_plot=False`, then you should call the
        `show()` method on the returned object in order to show the plots (which is simply
        an alias for `matplotlib.pyplot.show()`.
        """
        
        d = _Ns()

        d.fig = matplt.figure()

        d.ax = d.fig.add_subplot(111)

        d.ax.set_xlabel(xlabel)
        d.ax.set_ylabel('probability density')

        if log_scale:
            d.ax.set_yscale('log')

        f = self.fit_histogram_result.f
        p = self.fit_histogram_result.p
        errp = self.fit_histogram_result.errp

        fvals = np.linspace(np.min(f), np.max(f), curve_npts)

        d.ax.errorbar(x=f, y=p, yerr=errp, c='b', fmt='.', label='numerics')
        d.ax.plot(fvals, np.exp(self.fit_fn(self.ftox(fvals), *self.fit_params)), c='r', label='fit')
        if plot_deskewed_gaussian:
            if self.custom_fit_fn:
                raise RuntimeError("Cannot plot deskewed gaussian with a custom fit; use plot_deskewed_gaussian=False.")
            (a, x0, y0) = deskew_logmu_curve(self.fit_params.a2, self.fit_params.a1, self.fit_params.m, self.fit_params.c)
            d.ax.plot(fvals, np.exp(-a*(self.ftox(fvals)-x0)**2 + y0), c='g', label='deskewed Gaussian')

        if show_plot:
            matplt.show()
        else:
            d.show = matplt.show

        return d;
        
#


def load_tomorun_csv_histogram_file(fn):
    """
    Load a histogram data file produced by executing the `tomorun` executable program
    (instead of via the `tomographer.tomorun` python module).

    The argument `fn` is the file name. It is expected to be a TAB-separated CSV file
    where the first row (header) is blindly discarded.  Any fourth column is disregarded.

    Note: We blindly assume, without checking, that the first column is a list of linearly
    spaced values.

    The returned value is simply a :py:class:`tomographer.HistogramWithErrorBars` object.
    """
    import tomographer
    
    dat = np.loadtxt(fn, skiprows=1)
    # dat[:,0] should be lin-spaced values corresponding to the left edges of the histogram bins
    fmin = dat[0,0]
    numbins = dat.shape[0]
    binresolution = dat[1,0]-dat[0,0]
    fmax = dat[numbins-1,0] + binresolution
    h = tomographer.HistogramWithErrorBars(fmin, fmax, numbins)
    h.load(dat[:,1], dat[:,2])
    return h




"""
The convenience module `tomographer.include` facilitates the development of new
`Python`/`C++` modules using tomographer along with tomographerpy headers.
"""

import os.path

#
# The included components are defined here -- this present list is also used by setup.py
#
BOOST_DEPS_COMPONENTS = ['algorithm', 'math', 'core', 'exception', 'predef']
"""
Those components of boost which are included in the header dependencies directory of
this package.
"""


def get_include(keys=False):
    """
    Return the directory(ies) to add to C++ include path in order to compile
    extensions using `tomographer`/`tomographerpy` headers.  This function is
    intended to be used in your `setup.py` in order to compile your extension.

    Tomographer's dependencies, the Eigen and Boost headers, are included in the
    package data so that you can easily compile other extensions using
    Tomographer.  This function will also return the necessary include paths to
    reach those dependencies.  (Note: Only a subset of the boost headers are
    included, as defined by :py:data:`BOOST_DEPS_COMPONENTS`).

    If `keys=False` (the default), then a list of include paths are returned.
    If `keys=True`, then a dictionary is returned with the keys `['tomographer',
    'tomographerpy', 'boost', 'eigen']` and with corresponding values which are
    the include paths for each component.

    See also: Information about compile flags in
    :py:data:`tomographer.version.compile_info['cflags']
    <tomographer.version.compile_info>`.
    """

    root_include = os.path.dirname(os.path.realpath(__file__))

    if keys:
        return { 'tomographer': root_include,
                 'tomographerpy': root_include,
                 'boost': os.path.join(root_include, 'deps', 'boost'),
                 'eigen': os.path.join(root_include, 'deps', 'eigen3'), }
    
    return list(set(get_include(keys=True).values()))




def find_include_dir(headername, pkgnames=[], testfn=os.path.isdir, return_with_suffix=None, add_paths=[]):
    """
    Find an include directory for a third-party library.

    This function works by looking for `headername` using the test function `testfn`.  By
    default, it looks for a directory named `headername`.

    Arguments:

      - `headername`: file or directory name to look for
    
      - `pkgnames`: package names under which these headers may have been packaged, e.g.,
        by homebrew.  This is used to add guess paths.

      - `testfn`: how to determine that `headername` was found.  The default,
        :py:func:`os.path.isdir`, checks that `headername` is a directory.

      - `return_with_suffix`: you may set this to a path fragment to be appended to the
        directory found

      - `add_paths`: additional search paths to use

    Returns: The directory relative to which we found `headername`, possibly supplemented
    by the given suffix in `return_with_suffix`.  `None` is returned if the include files
    were not found.
    """
    guesses = (
        # homebrew version
        [ os.path.join('/usr/local/opt', pkgname, 'include') for pkgname in pkgnames] + [
            # default paths
            '/usr/local/include',
            '/usr/include',
            '/opt/include',
            os.path.expanduser('~/.local/include'),
        ] + add_paths)
    for i in guesses:
        dn = os.path.join(i, headername)
        if testfn(dn): # found
            if return_with_suffix:
                return os.path.join(i, return_with_suffix)
            return i

    return None


_staticlib_suffixes = ['.a']
_dylib_suffixes = ['.so', '.dylib', '.dll']

def find_lib(libname, pkgnames, libnamesuffixes=[], add_paths=[], prefer_static=True):
    """
    Find a third-party library.

    This function looks for the library `libname` using common library naming schemes
    (e.g. lib`libname`.dylib) in common search paths.

    Returns the full path to the library or `None` if not found.

    Arguments:

      - `libname`: the base name of the library, without any ``lib`` prefix or any suffix
    
      - `pkgnames`: as for :py:func:`find_include_dir`, specify possible package names
        which may contain this library

      - `libnamesuffixes`: possible library suffixes for `libname` (but not filename
        suffix). This is used for instance some times to distinguish a multithreaded
        version of a library from a non-multithreaded one)

      - `add_paths`: additional paths in which to look

      - `prefer_static`: if `True`, then static library suffixes are tried before the
        dynamic ones; if `False`, the opposite.
    """
    guesses = (
        # homebrew version
        [ os.path.join('/usr/local/opt', pkgname, 'include') for pkgname in pkgnames] + [
        os.path.join('/usr/local/opt', pkgname, 'lib'),
        # default paths
        '/usr/local/lib',
        '/usr/lib',
        '/opt/lib',
        '/usr/lib{suffix}/x86_64-linux-gnu',
        os.path.expanduser('~/.local/lib'),
    ] + add_paths)
    is_64bits = sys.maxsize > 2**32 # see https://docs.python.org/2/library/platform.html
    libdirsuffixes = ['', '64' if is_64bits else '32']
    for i in guesses:
        for s in libdirsuffixes:
            if '{suffix' not in i:
                i += '{suffix}'
            #print("find_lib:TRYING i={!r}".format(i))
            basedn = i.format(suffix=s)
            #
            for lnp in ['lib', '']: # lib-name-prefixes
                for lns in libnamesuffixes + ['']: # lib-name-suffixes
                    fnslist = (_staticlib_suffixes+_dylib_suffixes) if prefer_static else (_dylib_suffixes+_staticlib_suffixes)
                    for fns in fnslist: # file name suffixes for library
                        dn = os.path.join(basedn+s, lnp+libname+lns+fns)
                        #print("find_lib:TRYING dn={!r}".format(dn))
                        if os.path.isfile(dn):
                            return dn
    return None



def ensure_str(s):
    """
    Simple conversion of the value `s` to a `str`, that is, a byte string on Python 2 and
    a unicode string on Python 3.

    Input is expected to be a byte-string or unicode string in both Py2 or Py3.
    """
    if not isinstance(s, str):
        return s.decode('utf-8')
    return s



class Vars(object):
    """
    Minimal handling of compilation config variables, importable from CMake cache.

    Constructor arguments:
      - `vars`: the list of variable names which we will be storing.
      - `cachefile`: a CMake cache file to look values into.  If `None`, no CMake cache is
        consulted.

    The variables are looked up:
       - in the environment variables
       - in the CMake cache, if any is specified
       - as per any given default value (with :py:meth:`setDefault()`))

    Example usage::

        cmake_cache_file = os.environ.get('CMAKE_CACHE_FILE', None)
        vv = Vars([
            'GIT',
            'EIGEN3_INCLUDE_DIR',
        ], cachefile=cmake_cache_file)
        vv.setDefault('GIT', lambda : find_executable('git'))
        vv.setDefault('EIGEN3_INCLUDE_DIR',
                      lambda : find_include_dir('eigen3', pkgnames=['eigen','eigen3'],
                                                return_with_suffix='eigen3'))
    """
    def __init__(self, vars, **kwargs):
        # self.d: values read & cached
        self.d = {}

        dc = {}
        if 'cachefile' in kwargs:
            self._cachefile = kwargs['cachefile']
            if self._cachefile:
                dc = self._readcache(self._cachefile)

        for k in vars:
            if k in os.environ:
                self.d[k] = os.environ.get(k)
            elif k in dc:
                self.d[k] = ensure_str(dc[k])
            else:
                self.d[k] = None

    def setDefault(self, k, defvalue):
        """
        If no value has been detected for key `k`, set the value `defvalue`.  `defvalue` may
        be a callable which is invoked only if the value is actually needed.
        """
        if k not in self.d or self.d.get(k, None) is None:
            if callable(defvalue):
                self.d[k] = defvalue()
            else:
                self.d[k] = defvalue

    def _readcache(self, cachefile):
        dc = {}
        import re
        rx = re.compile(r'^(?P<name>[A-Za-z0-9_]+)(:(?P<type>[A-Za-z_]+))=(?P<value>.*)$')
        with open(cachefile) as f:
            for line in f:
                m = rx.match(line)
                if not m:
                    continue
                dc[m.group('name')] = ensure_str(m.group('value'))
        return dc

    def get(self, k):
        """
        Get the value associated with the variable `k` (specified in environment, read from
        cache, or default value).
        """
        return self.d.get(k)

    def message(self, tight=False):
        s = "Configuration variables:\n\n"
        s += ("\n".join([ "    {}={}".format(k,v) for k,v in self.d.items() ]))
        s += "\n\nYou may specify configuration variable values using environment variables."

        if hasattr(self, '_cachefile'):
            if not self._cachefile:
                s += ("\n\nYou may also read variables from a CMakeCache.txt file with\n"
                      "    CMAKE_CACHE_FILE=path/to/CMakeCache.txt")
            else:
                s += ("\n\n    (read cache file {})".format(cmake_cache_file))
        
        if tight:
            return s

        return "\n" + s + "\n"




"""
This module collects utilities gravitating around some standard tasks you may need to
perform when doing quantum tomography.

A routine allows you to simulate measurement outcomes from a true quantum state and a set
of measurement settings.  The standard Pauli measurement settings on a qubit are available
in :py:const:`~tomographer.tools.densedm.PauliMeasEffectsQubit`.

A separate submodule (:py:mod:`~tomographer.tools.densedm.mle`) determines the maximum
likelihood estimate given experimental data.

Further tools may be added to these modules in the future.
"""


import numpy as np


class _Ns: pass # dummy object acts as namespace

def simulate_measurements(rho, Mk, num_samples_per_setting):
    """
    Simulate measurements for quantum tomography.

    The "true state" `rho` should be a `numpy.array` object, as well as each POVM effect
    in `Mk`.

    `Mk` should be a list of lists of POVM effects, where ``Mk[k][i]`` is the POVM effect
    corresponding to outcome `i` of measurement setting `k`.  A POVM effect must be a
    positive semidefinite matrix (given as a `NumPy` array), and all effects of a
    measurement setting should sum up to the identity matrix.
    
    `num_samples_per_setting` specifies the number of repetitions of a measurement
    setting.  If it is an integer, then we simulate this number of measurement outcomes
    for each measurement setting in `Mk`.  If it is a list (or a `numpy.array`), then it
    should have the same length as there are settings in `Mk` (i.e. ``len(Mk) ==
    len(num_samples_per_setting)``), and for the `k`-th measurement setting in `Mk` we
    will simulate ``num_samples_per_setting[k]`` number of outcomes.

    Returns: an object `d` with properties `d.Emn` and `d.Nm`, representing the POVM
    effects and simulated frequency counts.  They are in a format suitable for input to
    :py:func:`tomographer.tomorun.tomorun()` or
    :py:class:`tomographer.densedm.IndepMeasLLH`.
    """

    num_settings = len(Mk)

    num_total_effects = np.sum([ len(Mk[k]) for k in range(len(Mk)) ])

    assert len(rho.shape) == 2 and rho.shape[0] == rho.shape[1]
    
    dim = rho.shape[0]

    # prepare the list of POVM effects, and simulate the measurements
    Emn = []
    Nm = []


    if (hasattr(num_samples_per_setting, "__getitem__")):
        # a list of numbers
        numsimoutcomes = lambda k: num_samples_per_setting[k]
    else:
        # just a single constant, use same number for all measurement settings
        numsimoutcomes = lambda k: num_samples_per_setting

    for k in range(num_settings):

        # get some random numbers for this measurement setting
        #print("DEBUG: numsimoutcomes=", numsimoutcomes(k))
        x = np.random.rand(numsimoutcomes(k))
        proboffset = 0

        # We sample the measurement outcomes as follows: we split the interval [0,1]
        # in number sections equal to the number of possible outcomes, each of length
        # = probability of that outcome.  Then, for each possible outcome, we count
        # the number of random numbers in `x` that fall into the corresponding
        # section.

        for i in range(len(Mk[k])):

            Ek = Mk[k][i]
            p = np.einsum('ij,ji->', Ek, rho) # = trace(Ek * rho)
            Emn.append( Ek.astype(dtype=complex) )
            Nm.append( np.count_nonzero( (proboffset <= x) & (x < proboffset+p) ) )

            proboffset += p

        # sanity check
        assert np.abs(proboffset-1.0) < 1e-4, (
            "POVM effects do not sum to identity or rho_sim not normalized, got proboffset={}".format(proboffset)
            )

    d = _Ns()
    d.Emn = Emn
    d.Nm = np.array(Nm)
    return d





#
# Measurement settings
#

PauliMeasEffectsQubit = [ [
        np.array([[.5, .5],[.5, .5]]),     # X, +1 outcome
        np.array([[.5, -.5],[-.5, .5]]),   # X, -1 outcome
    ], [
        np.array([[.5, -.5j],[.5j, .5]]),  # Y, +1 outcome
        np.array([[.5, .5j],[-.5j, .5]]),  # Y, -1 outcome
    ], [
        np.array([[1,0],[0,0]]),           # Z, +1 outcome
        np.array([[0,0],[0,1]]),           # Z, -1 outcome
    ]
] # docstring:
"""
List of POVMs corresponding to measuring Pauli operators on a single qubit.
``PauliMeasEffectQubit[i][s]`` is the POVM effect corresponding to measuring the outcome
indexed by `s` of the Pauli operator indexed by `i`.
"""




#from . import jpyutil as _jpyutil
#from . import querrorbars as _querrorbars
#def __bootstrap__():
#   global __bootstrap__, __loader__, __file__
#   import sys, pkg_resources, imp
#   __file__ = imp.find_module('_tomographer_cxx')[1]
#   __loader__ = None;
#   del __bootstrap__, __loader__
#   imp.load_dynamic(__name__, __file__)
#__bootstrap__()
#jpyutil = _jpyutil
#querrorbars = _querrorbars

# load the C++ module and import the relevant symbols here
from _tomographer_cxx import *

import _tomographer_cxx
__version__ = _tomographer_cxx.__version__

# set up docstrings for our classes
#from . import _docstrings





"""
Utility to find the MLE estimate, using `cvxpy <http://www.cvxpy.org/>`_.
"""


import numpy as np

import cvxpy

import scipy.sparse as sp

import tomographer.densedm


class _Ns: pass


DEFAULT_SOLVER_OPTS = {
    'solver': cvxpy.SCS,
    'eps': 1e-6,
    'max_iters': 5000,
    'verbose': True,
} # docstring:
"""
The default solver options used by :py:func:`find_mle()`.
"""


def find_mle(llh, solver_opts=None):
    r"""
    Find the Maximum Likelihood Estimate state.  The measurement data is specified in the
    `llh` argument as a :py:class:`tomographer.densedm.IndepMeasLLH` instance.

    This function creates and solves a `cvxpy` problem solving the following:

    .. math::

         \text{maximize:}\quad &\sum_k N_k \ln\operatorname{tr}(E_k\cdot\rho)

         \text{subject to:}\quad &\rho \geqslant 0

                            &\operatorname{tr}\rho = 1

    The return value of this function is an two-tuple `(rho_MLE, d)`, where `rho_MLE` is a
    `numpy.array` representing the maximum likelihood estimate, and where `d` is an object
    with the following attributes set:

    - ``d.result``: the raw return value of `cvxpy.solve()`

    - ``d.rho_MLE``: a copy of `rho_MLE`

    - ``d.rho_vars``: a two-tuple `(rho_R, rho_I,)` of the raw `cvxpy` variables used to
      represent the complex variable `rho`. (Cvxpy doesn't support complex variables, so
      we need to declare two real matrix variables, corresponding to the real and
      imaginary parts of the variable `rho`.)

    - ``d.objective``: the raw optimization objective object constructed for `cvxpy`

    - ``d.constraints``: the list of constraints specified to `cvxpy`

    - ``d.problem``: the raw `cvxpy` problem object

    The optional `solve_opts` argument is a dictionary of options that are directly passed
    on to `cvxpy`.  The keys specified in `solve_opts` override the default keys in
    :py:const:`DEFAULT_SOLVER_OPTS`.
    """

    dim = llh.dmt.dim
    dim2 = llh.dmt.dim2
    dimtri = (dim2 - dim) / 2

    if llh.numEffects == 0:
        logging.getLogger(__name__).warning("No measurements given, did you forget to populate the llh object?")
        return None

    rho_R = cvxpy.Symmetric(dim, dim)
    rho_I = cvxpy.Variable(dim, dim)

    constraints = [
        rho_I == -rho_I.T,
        cvxpy.vstack( cvxpy.hstack(rho_R, -rho_I) ,
                      cvxpy.hstack(rho_I,  rho_R) ) >> 0,
        cvxpy.trace(rho_R) == 1,
        ]

    # TODO: go via X parameterization for the optimization ?
    #

    px = tomographer.densedm.ParamX(llh.dmt)

    Nm = llh.Nx()
    Exn = llh.Exn()
    EmnR = []
    EmnI = []
    for k in range(len(Exn)):
        # bah, it's ugly to get these back into matrices, clean this up...
        Emn = px.XToHerm(Exn[k])
        EmnR.append(Emn.real)
        EmnI.append(Emn.imag)

    #print("DEBUG: EmnR = ", EmnR)
    #print("DEBUG: EmnI = ", EmnI)

    objective = cvxpy.Maximize( sum( [
        Nm[k] *
        cvxpy.log( cvxpy.trace(EmnR[k] * rho_R) - cvxpy.trace(EmnI[k] * rho_I) )
        for k in range(len(Exn))
    ]) )

    problem = cvxpy.Problem(objective, constraints)

    # default solver options
    final_solver_opts = dict(DEFAULT_SOLVER_OPTS)
    if solver_opts is not None:
        # include user-given options
        final_solver_opts.update(solver_opts)

    result = problem.solve(**final_solver_opts)

    d = _Ns()
    d.result = result
    d.rho_vars = (rho_R, rho_I,)
    d.rho_MLE = rho_R.value + 1j*rho_I.value
    d.objective = objective
    d.constraints = constraints
    d.problem = problem

    rho_MLE = d.rho_MLE

    if (problem.status != 'optimal'):
        rho_MLE = None
        logging.getLogger(__name__).warning("Error: couldn't solve the MLE problem. Status = "+str(problem.status))

    return (rho_MLE, d)




"""
Utilities for Jupyter notebooks.

Note that the utilities here have simple fallbacks to make them also work outside of
Jupyter notebooks.  For example, a :py:class:`SimpleProgressBar` would display a graphical
progress bar inside a Jupyter notebook, but inside a python script executed on the
console, it would simply print out progress information using ``print(...)``.
"""

try:
    from ipywidgets import IntProgress, HTML, VBox, HBox
    from IPython.display import display
except ImportError:
    pass # no problem, not in Jupyter/ipython notebook

try:
    import html
    escapehtml = html.escape
except ImportError:
    import cgi
    escapehtml = cgi.escape

class _Ns: pass

def _in_html_ipynb():
    #
    # Return True if we are in IPython with HTML display (e.g. in a jupyter notebook) --
    # required for progress bars
    #
    ipy = None
    try:
        # get IPython InteractiveShell instance
        ipy = get_ipython()
    except NameError:
        return False
    #
    # not sure which method is best (more future-proof). In any case, get the
    # DisplayFormatter object and...
    #
    # ... check if 'text/html' is in active_formatters ...
    try:
        if 'text/html' in ipy.display_formatter.active_formatters:
            return True
    except AttributeError:
        pass
    # ... or check that the formatter for the MIME type 'text/html' is "enabled"
    try:
        if ( 'text/html' in ipy.display_formatter.formatters  and
             ipy.display_formatter.formatters['text/html'].enabled ):
            return True
    except AttributeError:
        pass

    # by default, assume we're not in HTML-enabled IPython
    return False






class _SimpleProgressBar_JupyterImpl(object):
    def __init__(self, label_msg_html="Progress: ", displaybox=True):
        super(_SimpleProgressBar_JupyterImpl, self).__init__()
        self.progress_widget = IntProgress(min=0, max=100, value=0)
        self.label = HTML("<span style=\"padding-right: 1em;\">"+label_msg_html+"</span>")
        self.labelval = HTML()
        self.innerhbox = HBox(children=[self.label, self.progress_widget, self.labelval])
        self.addinfo = HTML()
        self.vbox = VBox(children=[self.innerhbox, self.addinfo])
        if displaybox:
            display(self.vbox)

    def __enter__(self):
        self.progress_widget.value = 0
        self.progress_widget.bar_style = 'info'
        return self

    def progress(self, fraction_done, addinfo=None):
        pcvalue = int(fraction_done*100)
        self.progress_widget.value = pcvalue
        self.labelval.value = "<span style=\"font-weight: bold; padding-left: 1em;\">{}% done</span>".format(pcvalue)
        self.addinfo.value = ("<pre style=\"display: block\">{}</pre>".format(escapehtml(addinfo)) if addinfo else "")

    def __exit__(self, etype, evalue, etraceback):
        if etype:
            self.progress_widget.bar_style = 'danger'
            self.labelval.value = "<span style=\"font-weight: bold; padding-left: 1em;\">(interrupted)</span>"
        else:
            self.progress_widget.value = 100
            self.progress_widget.bar_style = 'success'
            self.labelval.value = "<span style=\"font-weight: bold; padding-left: 1em;\">finished</span>"

class _SimpleProgressBar_ConsoleImpl(object):
    def __init__(self, label_msg_html="Progress: ", displaybox=True):
        super(_SimpleProgressBar_ConsoleImpl, self).__init__()
        self.addinfo = None

    def __enter__(self):
        return self

    def progress(self, fraction_done, addinfo=None):
        totw = 80 - 9;
        w = int(fraction_done*totw)
        print("|{}{}| {:5.2%}".format('>'*w, ' '*(totw-w), fraction_done))
        if addinfo:
            print(addinfo)

    def __exit__(self, etype, evalue, etraceback):
        pass


if _in_html_ipynb():
    _SimpleProgressBar_base = _SimpleProgressBar_JupyterImpl
else:
    # This also happens when parsed by Sphinx
    _SimpleProgressBar_base = _SimpleProgressBar_ConsoleImpl

#
# We need to subclass the relevant class here, because we want to include docstrings.
# Unfortunately with Python2 we cannot directly edit the docstrings as
# "SimpleProgressBar.__doc__ = ...", although that works in Python 3.
#
class SimpleProgressBar(_SimpleProgressBar_base):
    """
    A simple progress bar implementation, to be used in a Python `with` statement::

        with SimpleProgressBar("Working very hard ... ") as prg:
            for i in range(10000):
                prg.progress(fraction_done=i/10000.0)
                # do stuff that takes time
                ...


    This class works both inside a Jupyter notebook, providing a nice graphical progress bar,
    as well as in any other context such as a simple python console script, in which case
    information is simply displayed in the terminal using `print(...)`.

    In addition, it is possible to display additional information to simply the fraction done.
    You can do this by specifying the argument `add_info` to the :py:meth:`progress()`
    callback.  Inside a Jupyter notebook, the information is displayed below the progress bar
    and is constantly updated at each call of :py:meth:`progress()`.  In the console version,
    the information is simply displayed each time using ``print(...)``.

    Note: When we are inside a Jupyter notebook, there is an additional attribute which is
    exposed called `addinfo`.  It is an `IPython.display.HTML` widget which serves to display
    the additional info. By testing for this attribute you can display custom stuff in there,
    e.g. a final report after the work has finished.

    """
    
    def progress(self, *args, **kwargs):
        """
        progress(fraction_done[, add_info=None])

        The callback function to update the progress bar (or to display new information on the
        console).

        :param fraction_done: is the fraction of the job done, a real number between zero and one.
        :param add_info: any additional information (formatted as plain text) which should be
            displayed alongside the progress bar. If `None` (or not specified), then no additional
            information is displayed.

        """
        super(SimpleProgressBar, self).progress(*args, **kwargs)






class RandWalkProgressBar(SimpleProgressBar):
    """
    A progress bar suitable for displaying the status of random walk tasks (as for
    :py:func:`tomographer.tomorun.tomorun()`, for example).

    This class automatically provides a callback function :py:meth:`progress_fn` which
    accepts a :py:class:`tomographer.multiproc.FullStatusReport`, and is capable of
    displaying meaningful additional information alongside the progress bar.  The callback
    :py:meth:`progress_fn` is suitable to be specified directly to, e.g.,
    :py:func:`tomographer.tomorun.tomorun()`.

    Example::

        result = None
        with RandWalkProgressBar() as prg:
            result = tomographer.tomorun.tomorun(..., progress_fn=prg.progress_fn)
            prg.displayFinalInfo(result['final_report_runs'])

    This class works both inside a Jupyter notebook as well as in other contexts such as a
    console for a simple Python script.

    .. note:: The callback method which should be called regularly is `progress_fn()`, and
              not `progress()`.
    """
    def __init__(self, label_msg_html="Random Walk Progress: "):
        super(RandWalkProgressBar, self).__init__(label_msg_html=label_msg_html)

    def progress(self, *args, **kwargs):
        """
        Do not call this method. It will raise a `RuntimeError`.
        """
        raise RuntimeError("Set as callback RandWalkProgressBar.progress_fn, not RandWalkProgressBar.progress")

    def progress_fn(self, report):
        """
        Update the progress bar and additional info to display the current status of the
        random walk specified in `report`, provided as a
        :py:class:`~tomographer.multiproc.FullStatusReport` object.

        This method is suitable to be specified as a callback to, e.g.,
        :py:func:`tomographer.tomorun.tomorun()`.
        """
        super(RandWalkProgressBar, self).progress(
            fraction_done=report.total_fraction_done,
            addinfo=report.getHumanReport()
            )

    def displayFinalInfo(self, report_str):
        """
        Display some text in the "additional information" area of the progress bar.

        This is useful if you want to display some final report after the random walks
        have completed.

        This method works both within a Jupyter notebook as well as in a console.
        """
        if self.addinfo:
            self.addinfo.value += "<pre style=\"display: block\">{}</pre>".format(escapehtml(report_str))
        else:
            print(report_str)






# This file is part of the Tomographer project, which is distributed under the
# terms of the MIT license.
#
# The MIT License (MIT)
#
# Copyright (c) 2016 ETH Zurich, Institute for Theoretical Physics, Philippe Faist
# Copyright (c) 2017 Caltech, Institute for Quantum Information and Matter, Philippe Faist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from distutils.spawn import find_executable
from distutils.version import LooseVersion

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

import codecs
import os
import os.path
import sys
import subprocess # for git
import re
import glob
import shutil
import shlex

# build-time requirements -- Apparently, PIP cannot install these automatically -- so make
# sure we get a hard failure if they aren't present
try:
    import numpy
    import pybind11
except ImportError as e:
    sys.stderr.write(
        "\n"
        "*** ERROR: Please install packages `numpy' and `pybind11' prior to installing `tomographer' ***\n"
        "    [for instance: pip install numpy pybind11 ]\n"
        "\n")
    raise

def require_mod_version(mod, modver, minver, fix=None):
    if LooseVersion(modver) < LooseVersion(minver):
        msg = "Minimum required {mod} version is {minver}".format(mod=mod,minver=minver)
        if fix:
            msg += "; try {}".format(fix)
        raise RuntimeError(msg)


# Not too old NumPy
require_mod_version("NumPy", numpy.__version__, "1.8",
                    fix='running "pip install numpy --upgrade"')

# Very recent Pybind11, need >= 2.1 for automatic py::metaclass()
require_mod_version("PyBind11", pybind11.__version__, "2.1",
                    fix='running "pip install pybind11 --upgrade"')



#
# POSSIBLE ENVIRONMENT VARIABLES WHICH CAN BE SET: -- doc printed on standard output anyway
#
# handled natively by setup.py:
# CC=/path/to/compiler/gcc
#
# handled by us: (can be found in CMake cache)
# GIT=/path/to/git
# Boost_INCLUDE_DIR=/path/to/boost/headers
# EIGEN3_INCLUDE_DIR=/path/to/eigen/headers
# PYBIND11_CPP_STANDARD=-std=c++14
#
# handled by us (not in CMake cache):
# BCP=/path/to/bcp  (Boost packaging tool)
# OPTIMIZATION_CXX_FLAGS=-UNDEBUG -march=native -O3
#                        (compiler SIMD instruction sets & other optimizations)
#
# To read the above variables from cache:
# CMAKE_CACHE_FILE=path/to/CMakeCache.txt
#




thisdir = os.path.dirname(os.path.realpath(os.path.abspath(__file__)))
print("Tomographer py dir = {}; cwd = {}".format(thisdir, os.getcwd()))



sys.path.insert(0, os.path.join(thisdir, 'tomographer'))
# tomographer.include:
from include import find_include_dir, find_lib, Vars, ensure_str   # beurk!!

            



cmake_cache_file = os.environ.get('CMAKE_CACHE_FILE', None)
vv = Vars([
    'GIT',
    'BCP',
    'Boost_INCLUDE_DIR',
    'EIGEN3_INCLUDE_DIR',
    'PYBIND11_CPP_STANDARD',
    'OPTIMIZATION_CXX_FLAGS',
    'CXX_FLAGS',
#    'MACOSX_CXX_FLAGS',
], cachefile=cmake_cache_file)

# Defaults: programs (git, bcp)
vv.setDefault('GIT', lambda : find_executable('git'))
vv.setDefault('BCP', lambda : find_executable('bcp'))

# Defaults: flags
#
# DO NOT USE -fvisibility=hidden OR -flto. (Event with visibility attributes set I
# couldn't get other modules which use the python tomographer API to load properly)
vv.setDefault('OPTIMIZATION_CXX_FLAGS', '-UNDEBUG -march=native -O3')

# Defaults: Boost stuff
vv.setDefault('Boost_INCLUDE_DIR',
              lambda : find_include_dir('boost', pkgnames=['boost']))

# Defaults: Eigen3
vv.setDefault('EIGEN3_INCLUDE_DIR',
              lambda : find_include_dir('eigen3', pkgnames=['eigen','eigen3'],
                                        return_with_suffix='eigen3'))

#
# Default: C++ flags for Mac OS X
#
def dflt_cxxflags():
    if sys.platform == 'darwin':
        return "-stdlib=libc++"
    return ""
vv.setDefault('CXX_FLAGS', dflt_cxxflags)



#
# PREPARE THE PYTHON SOURCES
#

import setup_sources

IsTomographerSources = setup_sources.setup_sources(thisdir, vv)

print("IsTomographerSources = {!r}".format(IsTomographerSources))

#
# Read version info
#
with open(os.path.join(thisdir, 'VERSION')) as f:
    version = ensure_str(f.read()).strip()

(version_maj, version_min, version_for_pip) = setup_sources.get_version_info(version)



#
# Tell the user everything.
#

print("""
  Welcome to the setup.py script for Tomographer {version} ({version_for_pip}).
""".format(version=version, version_for_pip=version_for_pip))

print("""\
  The `tomographer` python package requires some external C++ libraries. You may
  need to specify their location with the use of environment variables. Current
  values (detected or specified manually) are:

{}""".format("\n".join([ "    {}={}".format(k,v) for k,v in vv.d.items() ])))

if IsTomographerSources and not cmake_cache_file:
    print("""
  You may also read variables from a CMakeCache.txt file with
  CMAKE_CACHE_FILE=path/to/CMakeCache.txt
""")
else:
    print("""
  (read cache file {})
""".format(cmake_cache_file))


if IsTomographerSources and sys.platform == 'darwin':
    print("""\
  To get started on Mac OS X with homebrew, you may run for instance:

    > brew install eigen3 boost pybind11
""")


#
# Set up the compilation options
#

glob_cflags = [
    # no longer needed -- now we create tomographer_version.h
    # '-DTOMOGRAPHER_VERSION=\"{}\"'.format(version),
    # '-DTOMOGRAPHER_VERSION_MAJ={}'.format(version_maj),
    # '-DTOMOGRAPHER_VERSION_MIN={}'.format(version_min),
    '-DEIGEN_DONT_PARALLELIZE',
]

cxxfiles = [ os.path.join('cxx', x) for x in [
    "tomographerpy.cxx",
    "pyhistogram.cxx",
    "pymultiproc.cxx",
    "pymhrwtasks.cxx",
    "pymhrw.cxx",
    "pydensedm.cxx",
    "pytomorun.cxx",
] ]

headers = [ os.path.join('tomographer/include', 'tomographerpy', x) for x in [
    "pyhistogram.h",
    "pymultiproc.h",
    "pymhrwtasks.h",
    "pymhrw.h",
    "pylogger.h",
    "pydensedm.h",
    "common.h"
] ]



#
# Some code taken from https://github.com/pybind/python_example
#

pybind_includes = [
    pybind11.get_include(False),
    pybind11.get_include(True),
]
numpy_includes = [
    numpy.get_include()
]




# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.
    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


def escape_c_utf8_str(s):
    _rx_notok = re.compile(r"""[^a-zA-Z_0-9.-/<>?~!@#$%^&*\[\]()_+={};':" -]""")
    def esc_char(x):
        if not isinstance(x, int):
            x = ord(x)
        if not _rx_notok.match(chr(x)):
            return chr(x)
        return '\\x%02x'%(x)
    return ('"' +
            "".join([ esc_char(x) for x in ensure_str(s).encode('utf-8') ])
            + '"')

class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])

        def defpp(sym, val=None):
            s = sym
            if val is not None:
                s += '='+str(val)
            if ct == 'msvc':
                return '/D'+s
            # -DSYMBOL=Value
            return '-D'+s

        stdcpp = vv.get('PYBIND11_CPP_STANDARD')
        if stdcpp:
            opts.append(stdcpp)
        else:
            opts.append(cpp_flag(self.compiler))

        for cflg in glob_cflags:
            opts.append(cflg)

        for cflg in shlex.split(vv.get('CXX_FLAGS')):
            opts.append(cflg)

        for cflg in shlex.split(vv.get('OPTIMIZATION_CXX_FLAGS')):
            opts.append(cflg)

        # create a header so that the C++ files have access to some setup.py info
        if not os.path.exists(os.path.join(thisdir, 'tmp')): os.mkdir(os.path.join(thisdir,'tmp'))
        with open(os.path.join(thisdir,'tmp','tomographerpy_compileinfo.h'), 'w') as f:
            f.write("""
#pragma once
#include <vector>
#include <string>

static inline std::vector<std::string> tomographerpy_compileinfo_get_cflags() {
    return { %(cflagslist)s } ;
}

static inline py::dict tomographerpy_compileinfo_get_compiler() {
    py::dict d;
    d["type"_s] = std::string(%(compiler_type)s);
    d["path"_s] = std::string(%(compiler_path)s);
    d["link_path"_s] = std::string(%(compiler_link_path)s);
    return d;
}
""" % {
    'cflagslist': ", ".join([ 
        'std::string(' + escape_c_utf8_str(opt) + ')'
        for opt in opts
    ]),
    'compiler_type': escape_c_utf8_str(self.compiler.compiler_type),
    'compiler_path': escape_c_utf8_str(self.compiler.compiler_so[0]),
    'compiler_link_path': escape_c_utf8_str(self.compiler.linker_so[0]),
}
            )
            opts.append(defpp("TOMOGRAPHERPY_HAVE_COMPILEINFO"))
            opts.append("-I"+os.path.join(thisdir,'tmp'))

        # don't include VERSION_INFO in the compileinfo header file
        opts.append(defpp('VERSION_INFO', '"%s"'%self.distribution.get_version()))
        opts.append(defpp('_tomographer_cxx_EXPORTS'))

        for ext in self.extensions:
            ext.extra_compile_args = opts

        build_ext.build_extensions(self)


def find_all_files(dirname, pred, prefixfn=lambda root, filename: os.path.join(root, filename)):
    matches = []
    for root, dirnames, filenames in os.walk(dirname):
        matches += [ prefixfn(root, filename) for filename in filenames
                     if pred(filename) ]
    return matches


target_tomographer_include = os.path.join(thisdir, 'tomographer', 'include')
target_tomographer_include_deps = os.path.join(target_tomographer_include, 'deps')

tomographer_include_files = (
    find_all_files(os.path.join(target_tomographer_include, 'tomographerpy'),
                   pred=lambda fn: fn.endswith('.h'),
                   prefixfn=lambda root, fn: os.path.relpath(os.path.join(root, fn), target_tomographer_include)) +
    find_all_files(os.path.join(target_tomographer_include, 'tomographer'),
                   pred=lambda fn: True,
                   prefixfn=lambda root, fn: os.path.relpath(os.path.join(root, fn), target_tomographer_include)) +
    find_all_files(os.path.join(target_tomographer_include_deps, 'boost', 'boost'),
                   pred=lambda fn: True,
                   prefixfn=lambda root, fn: os.path.relpath(os.path.join(root, fn), target_tomographer_include)) +
    find_all_files(os.path.join(target_tomographer_include_deps, 'eigen3'),
                   pred=lambda fn: True,
                   prefixfn=lambda root, fn: os.path.relpath(os.path.join(root, fn), target_tomographer_include))

)

#print("ALL INCLUDE FILES = {}".format(tomographer_include_files))




def readfile(x):
    with open(os.path.join(thisdir, x)) as f:
        return f.read()


#
# Build & installation requirements
#
REQUIREMENTS = [
    # recent PIP is required (get errors otherwise)
    'pip >= 7.1',
    'numpy >= 1.8',
    'pybind11 >= 2.1',
],


#
# Set up the tomographer python package
#

url = 'https://github.com/Tomographer/tomographer/'
download_url = ('https://github.com/Tomographer/tomographer/releases/download/'
                '{version}/tomographer-{version}.tar.gz').format(version=version)

setup(
    #
    # Essential meta-information
    #
    name="tomographer",
    version=version_for_pip,
    description=u'The Tomographer Project \u2014 Practical, Reliable Error Bars in Quantum Tomography',
    long_description=readfile('README.rst'),
    author='Philippe Faist',
    author_email='phfaist@caltech.edu',
    license='MIT',
    url=url,
    download_url=download_url,
    keywords='quantum tomography error bars',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: C++',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Physics',
    ],

    #
    # Included pure-Python tomographer packages
    #
    packages=[
        'tomographer',
        'tomographer.tools',
        'tomographer.tools.densedm',
        'tomographer.include',
    ],

    #
    # Our C++/PyBind11 extension module.
    #
    ext_modules=[
        Extension(
            '_tomographer_cxx',
            sources=cxxfiles,
            include_dirs=(
                # Path to pybind11 headers
                pybind_includes + 
                numpy_includes +
                # our own internal header dependencies
                [
                os.path.join(target_tomographer_include), # tomographer & tomographerpy
                os.path.join(target_tomographer_include_deps, 'boost'),
                os.path.join(target_tomographer_include_deps, 'eigen3'),
                ]),
            depends=headers,
            language='c++',
        ),
    ],
    
    #
    # package data - include our headers inside the package, so that other extensions
    # can depend easily on the tomographer header library (along with Eigen & Boost
    # dependencies)
    #
    package_data={
        'tomographer.include': tomographer_include_files,
    },
    
    #
    # Installation requirements
    #
    install_requires = REQUIREMENTS,
    #
    # Beware of setup_requires --
    # https://pip.pypa.io/en/stable/reference/pip_install/#controlling-setup-requires
    #
    # But setup_requires would apparently be the only solution for us, as PIP won't
    # install the install_requires dependencies before building the wheel :( -- but still
    # doesn't work, not sure why :( :(  -- Arrhgggh!!!
    #
    # Just document dependencies.  Apparently PIP doesn't support doing this
    # automatically.
    #
    #setup_requires = REQUIREMENTS,
    
    #
    # Our customized build procedure
    #
    cmdclass={ 'build_ext': BuildExt },
    
    #
    # Not safe for keeping in a ZIP file (because of our extension)
    #
    zip_safe=False,

)
